#include "arith.h"

/* longlist member functions  */

longlist::longlist(long n) {length=n; items=new long[n];}
longlist::~longlist() {length=0; delete[] items;}  

longlist::longlist(const longlist& l) 
{
  length=l.length;
  items=new long[length]; 
  long n=length; long* x=items, *y=l.items;
  while(n--) {*x++=*y++;}
}

longlist::longlist(const longlist& l, long n) 
{
  length= n<=l.length ? n : l.length;
  items=new long[length]; 
  n=length; long* x=items, *y=l.items;
  while(n--) {*x++=*y++;}
}

longlist& longlist::operator=(const longlist&l)
{
  if (this==&l) return *this;
  delete[] items; items=new long[length=l.length]; 
  long n=length; long* x=items, *y=l.items;
  while(n--) {*x++=*y++;}
  return *this;
}

long longlist::locate(long num) const
{
  for (long i=0; i<length; i++)  if (num == items[i]) return i;
  return -1;
}

void longlist::output(ostream& os) const
{
  long i=0;long *it = items;
  while(i<length) { if(i++)os<<", ";  os<<(*it++);}
}

int startswith(const longlist& a, const longlist& b, long l, long from)
{
  int ans=1;
  for (long i=from; (i<l)&&ans; i++) ans = (a(i)==b(i));
  return ans;
}

/* Prime number procs; adapted from Pari  */

byteptr diffptr;
long NPRIMES, BIGGESTPRIME;

long initprimes(long maxnum)  /* initializes global variable diffptr */
                             /* to give primes up to maxnum, returning
                                the number of primes  */
{
  register long k,size=(maxnum+257)>>1;
  byteptr p=(byteptr)calloc(size,1);
  register byteptr q,r,s,fin=p+size-1;
  if (!p) {cerr<<"Out of memory in initprimes!"<<endl;exit(1);}
  for(r=q=p,k=1;r<fin;)
    {
      do {r+=k; k+=2; r+=k;} while (*++q);
      for(s=r;s<fin;s+=k) *s=1;
    }
  r=p; *r++=2; *r++=1;
  for(s=q=r-1;;)
    {
      while (*++q);
      if (q>=fin) break;
      *r++=(q-s)<<1;
      s=q;
    }
  *r++=0;
  NPRIMES=r-p-1;
  BIGGESTPRIME = 0;
  if (diffptr = (byteptr)realloc(p,r-p)) 
    {
      BIGGESTPRIME = prime_number(NPRIMES);
      return NPRIMES;  
    }
  else return 0;
}

long prime_number (long n)  /* returns n'th prime from list */
{
  byteptr p=diffptr;
  long c,pr=0;
  while (n--) if (c= *p++) pr += c; 
              else /* error */
                   {cerr<<"Not enough primes!"<<endl;exit(1);}
  return pr;
}

longlist primes (long n)  /* returns list of first n primes */
{
  longlist ans(n);
  byteptr p=diffptr;
  long diff,pr=0;
  for (long i=0; i<n; i++)
    {
      if (diff= *p++) pr+=diff; 
      else /* error */
           {cerr<<"Not enough primes!"<<endl;exit(1);}
      ans.set(i,pr);
    }
  return ans;
}

long primdiv(long aa)
{
 primevar pr;
 long p=0,q;
 long a = abs(aa);
 while (pr.ok() && p==0)
   {q=pr; pr++;
    if (a%q==0) p = q;
    else if (q*q>a) p=a;   // N.B. this causes a=1 to return 1.  Beware!
   }
 if (p==0) {p=a; 
            cerr<<"No prime divisor found for "<<aa<<" so assuming prime!\n";
           }
 return p;
}


longlist pdivs(long aa)
{longlist plist(9);   // max number of prime factors of a long (<2^31)
 long np = 0;
 primevar pr; long a = abs(aa);long p;
 while ( (a>1) && (pr.ok()))
 { p = pr; pr++;
   if (a%p==0) {plist.set(np++, p);
                while (a%p==0) a/=p;      //divide out by all the p's in a
               }
   else if (p*p>a) {plist.set(np++, a);   a=1;
                   }
 }
 if (a>1) {plist.set(np++,a);}  //In case of p-factors outside range, assume
                                //the cofactor is prime.
 plist.truncate(np);
 return plist;
}


longlist posdivs(long a, const longlist& plist)
{
// cout << "In posdivs with a = " << a << endl;
 long np = plist.length;
// cout << np << " primes: " <<endl; cout << plist << endl;
 long nd = 1;
 longlist dlist(512);
 dlist.set(0, 1);
// cout << "Divisor 0 = 1" << endl;
 long p;
 for(longvar pr(plist); pr.ok(); pr++)
 {
  p = pr; long e = val(p,a);
  for (long j=0; j<e; j++)
  {
   for (long k=0; k<nd; k++)
   {
    dlist.set(nd*(j+1)+k , p*dlist[nd*j+k]);
//    cout << "Divisor " << (nd*(j+1)+k) << " = " << (p*dlist[nd*j+k]) << endl;
   }
  }
  nd*=(e+1);
 } 
 dlist.truncate(nd);
 return dlist;
}

longlist alldivs(long a, const longlist& plist)
{//cout << "In alldivs with a = " << a << endl;
 long np = plist.length;
// cout << np << " primes: " <<endl; cout << plist << endl;
 long nd = 2;
 longlist dlist(512);
 dlist.set(0, 1);
 dlist.set(1,-1);
// cout << "Divisor 0 =  1" << endl;
// cout << "Divisor 1 = -1" << endl;
 long p;
 for(longvar pr(plist); pr.ok(); pr++)
 {
  p = pr; long e = val(p,a);
//  cout << "p = " << p << ", e = " << e << endl;
  for (long j=0; j<e; j++)
  {
   for (long k=0; k<nd; k++)
   {
    dlist.set(nd*(j+1)+k , p*dlist[nd*j+k]);
//    cout << "Divisor " << (nd*(j+1)+k) << " = " << (p*dlist[nd*j+k]) << endl;
   }
  }
  nd*=(e+1);
 }
 dlist.truncate(nd);
 return dlist;
}

longlist sqdivs(long a, const longlist& plist)
{
 long np = plist.length;
 long nd = 1;
 longlist dlist(512);
 dlist.set(0, 1);
 long p;
 for(longvar pr(plist); pr.ok(); pr++)
 {
  p = pr; long e = val(p,a)/2;
  for (long j=0; j<e; j++)
  {
   for (long k=0; k<nd; k++)
   {
    dlist.set(nd*(j+1)+k , p*dlist[nd*j+k]);
   }
  }
  nd*=(e+1);
 }
 dlist.truncate(nd);
 return dlist;
}

longlist sqfreedivs(long a, const longlist& plist)
{
 long np = plist.length;
 long nd = 1;
 longlist dlist(512);
 dlist.set(0, 1);
 long p;
 for(longvar pr(plist); pr.ok(); pr++)
 {
  p = pr; long e = 1;
  for (long j=0; j<e; j++)
  {
   for (long k=0; k<nd; k++)
   {
    dlist.set(nd*(j+1)+k , p*dlist[nd*j+k]);
   }
  }
  nd*=(e+1);
 }
 dlist.truncate(nd);
 return dlist;
}


long mod(long a, long b)
{long c;
 if (b<0) b=-b;
 if (a>=0) c=a%b; else c=b-((-a)%b);
 if (c>(b>>1)) c-=b;
 return(c);
}


long posmod(long a, long b)
{
  long c=a%b;  
  if (c<0) return(c+b);
  return(c);
}


long bezout(long aa, long bb, long& xx, long& yy)
{long a,b,c,x,oldx,newx,y,oldy,newy,q;
 oldx = 1; oldy = 0; x = 0; y = 1; a = aa; b = bb;
 while (b!=0)
 { q = a/b; 
   c    = a    - q*b; a    = b; b = c;
   newx = oldx - q*x; oldx = x; x = newx;
   newy = oldy - q*y; oldy = y; y = newy;
  }
 if (a<0) {xx=-oldx; yy=-oldy; return(-a);}
 else     {xx= oldx; yy= oldy; return( a);}
}
 
long invmod(long a, long p)
{long g,x,y;
 g=bezout(a,p,x,y);
 if (g==1) return x;
 else {cerr << "invmod called with " << a << " and " << p << " -- not coprime!\n";
       return 0;
      }
}


int modrat(long n, long m, float lim, long& a, long& b)
{long q,r,t,qq,rr,tt,quot; int sucess;
// cout<<"modrat("<<n<<","<<m<<")";
 q=m; r=posmod(n,m); qq=0; rr=1; t=0; tt=0; a=r; b=1; 
 if (r<lim) 
   { 
//   cout<<" = "<<a<<"/"<<b<<"\n";
     return 1;
   }
 while (r!=0) 
 { 
   quot = q/r;
// cout<<"q,r,t = "<<q<<" "<<r<<" "<<t<<"\n";
   t  =  q-quot*r;   q = r;   r = t;
   tt = qq-quot*rr; qq = rr; rr = tt;
   if (r<lim)
     {
       if (abs(rr)<lim) {a=r; b=rr; return 1;}
       else cerr << "modrat error: no reconstruction for " << n << " mod " << m << "\n";
//     cout<<" = "<<a<<"/"<<b<<"\n";
       return 0;
     }
 }
 cerr << "modrat error: common factor with " << n << " mod " << m << "\n";
 return 0;
}


long val(long factor, long number)
{
 long n = abs(number), f = abs(factor);
 if ((n==0) || (f<2)) return 99999;  // error condition! N.B. This value 
                                     // must be unlikely and POSITIVE.
 long e = 0;
 while (n%f==0) {e++; n/=f;}
 return e;
}




int intbezout(int aa, int bb, int& xx, int& yy)
{int a,b,c,x,oldx,newx,y,oldy,newy,q;
 oldx = 1; oldy = 0; x = 0; y = 1; a = aa; b = bb;
 while (b!=0)
 { q = a/b; 
   c    = a    - q*b; a    = b; b = c;
   newx = oldx - q*x; oldx = x; x = newx;
   newy = oldy - q*y; oldy = y; y = newy;
  }
 if (a<0) {xx=-oldx; yy=-oldy; return(-a);}
 else     {xx= oldx; yy= oldy; return( a);}
}

long chi2(long a)
{ static long table8[8] = {0,1,0,-1,0,-1,0,1};
  return table8[posmod(a,8)];
}
 
long chi4(long a)
{ static long table4[4] = {0,1,0,-1};
  return table4[posmod(a,4)];
} 

long hilbert(long a, long b)
{ static long table44[4][4] = {{0,0,0,0},
                              {0,1,0,1},
                              {0,0,0,0},
                              {0,1,0,-1}};
  return table44[posmod(a,4)][posmod(b,4)];
}

long leg(long a, long b)  //nb this function is not intended for public use
{ long aa = a;
  long bb = b;
  long ans = 1;
  while (bb>1) 
  {     aa = aa % bb;
        if (aa<0)       {aa=-aa; ans*=chi4(bb);}
        while (!(aa%4)) {aa/=4;}
        if (!(aa%2))    {aa/=2; ans *= chi2(bb);}
        ans*=hilbert(aa,bb);
        long c=bb; bb=aa; aa=c;
  }
  return ans;
}
 
long legendre(long a, long b)
{ 
  return (((gcd(a,b)==1) && (b%2)) ? leg(a,b) : 0);
}
 
long kronecker(long d, long n)
{ long ans=0; long m=n, d4=d%4; if(d4<0)d4+=4;
  if ((gcd(d,n)==1) && ((d4==0)||(d4==1)))
  { ans=1;
    while (!(m%4)) m/=4;
    if (!(m%2)) {m/=2; ans*=((d-1)%8 ? -1 : 1);}
    ans *= legendre(d,m);
  }
  return ans;
}

long log2(long n,long& e)       /* returns 1,e if n=2^e, else 0 */
{ e = 0;
  if (n<1) return 0;
  while (!(n%2)) {n/=2; e+=1;}
  return (n==1);
}

/* END OF FILE */
