// MRANK1.CC: two-descent with quartic search
//
//
#include "points.h"   // from qcurves library
#include "log2.h"
#include "mquartic.h"
#include "mequiv.h"
#include "msoluble.h"
#include "qc.h"
#define USE_BIGINTS
#include "mrank1.h"

#ifdef USE_BIGINTS
//#define DEFAULT_NAUX 12
#define DEFAULT_NAUX 6
#else
#define DEFAULT_NAUX 5
#endif

//#define NO_PADIC_FILTERING

//#define SHOW_ABC_RANGES
//#define DEBUG_AH

#define maxnquartics 2048 // OK for curves of rank 12 or less.
#define abceps 0.001      // used in abc-test

#define ROUNDADJUST 0.001 // Amount add/subtracted before rounding for safety
long roundtemp;
#define ROUNDUP(x) (Iasb(roundtemp,ceil((x)-ROUNDADJUST)),roundtemp)
#define ROUNDDOWN(x) (Iasb(roundtemp,floor((x)+ROUNDADJUST)),roundtemp)

#define NEQPLIST 5        // Number of primes for equiv-test sieving

#ifndef USE_BIGINTS
int xsqrt(bigfloat a, bigfloat &b)
{
  if(a<0) return 0;
  b=floor(sqrt(a)+0.1);
  return is_zero(b*b-a);
}
#endif

#ifdef LiDIA
long lfloor(bigfloat x) {return  I2long(Ifloor(x));}
long lround(bigfloat x) {return  I2long(Iround(x));}
long lceil(bigfloat x)  {return  I2long(Iceil(x));}
#else
long lfloor(bigfloat x) {return  (long)floor(x);}
long lround(bigfloat x) {return  (long)floor(x+0.5);}
long lceil(bigfloat x)  {return -(long)floor(-x);}
#endif

void rank1::addquartic() // process latest quartic found
{
  long firsti, i, j, oldnumber, thisnumber, nfl; unsigned long code;  
  char ab;
  int trivial=0, newone=1, gls=0, els=0;
  quartic * qlist, *thisq;;
  bigint x,y,z, badp; 	  Point Ptemp;
  int atype = (ipivot<0)&&(type>1);

  if(atype) 
    {qlist=qlista; thisnumber=nquarticsa; nfl=nfirstlota; ab='A';}
  else 
    {qlist=qlistb; thisnumber=nquarticsb; nfl=nfirstlotb; ab='B';}
  thisq=qlist+thisnumber;

  if (verbose) cout << (*thisq) << "\t";

  // Check triviality

  if(atype) trivial = thisq->trivial(); // else certainly nontrivial
  if (trivial)
    {
      if (verbose) cout << "--trivial"<<endl;
      return;
    }
  if (verbose) cout<<"--nontrivial..."<<flush;

  // Check current is inequivalent to previous

  code = thisq->set_equiv_code(eqplist);
  firsti = (extra2==1 ? nfl : 0);
  for (i=firsti; newone && (i<thisnumber); i++)
    { 
      if((!atype)&&(!qlistbflag[i])) continue;
      if(traceequiv) cout << "\nTesting equiv with number " << ab<< i+1 << endl;
#ifdef NEW_EQUIV
      newone = ! new_equiv(thisq,qlist+i,traceequiv);
#else
      newone = !     equiv(thisq,qlist+i,dlist,traceequiv);
#endif
      if (!newone) oldnumber=i+1;
    }

  if(newone)      // Check local and global solubility:
    {
      if(atype) 
	{
	  nquarticsa++;  
	  if(verbose) cout<<"--new (A) #"<<nquarticsa<<"\t"<<flush;
	}
      // but we do not increment nquarticsb unless the quartic is
      // els with no point, as otherwise we do not need to keep it

      if (ratpoint(*thisq,1,lim1,x,y,z))
	{ 
	  gls=els=1;
	  if (verbose) cout<<"(x:y:z) = ("<<x<<" : "<<y<<" : "<<z<<")\n";
	}
      else
	{
	  if (locallysoluble(*thisq,plist,badp))
	    { 
	      els=1;
	      if (verbose) cout<<"locally soluble..."<<flush;
	      quartic_sieve qs(thisq,2,0);
	      if(qs.search(lim2))
		{ 
		  qs.getpoint(x,y,z); gls=1;
		  if (verbose) 
		    cout<<"(x:y:z) = ("<<x<<" : "<<y<<" : "<<z<<")\n";
		}
	      else 
		if (verbose) 
		  cout<<"no rational point found (limit "<<lim2<<")"<<flush;
	    }
	  else 
	    if (verbose) 
	      cout<<"not locally soluble (p = "<<badp<<")\n";
	}
      
      if(gls)
	{
  //cout<<"Calling qc() with (x:y:z) = ("<<x<<" : "<<y<<" : "<<z<<")\n";  
	  qc(*thisq,x,y,z,the_curve,&IJ_curve,tr_u,tr_r,tr_s,tr_t, Ptemp,verbose);
  //cout<<"qc() returns giving point " << Ptemp << "\n";
	  if(atype)
	    {
	      pointlist1[npoints1]=Ptemp;
	      npoints1++; 
	      n1++; n2++;
	      if(verbose) cout<<"Size of A=ker(eps) increases to "<<n1<<endl;
	    }
	  else
	    {
	      pointlist2[npoints2]=Ptemp;
	      npoints2++; 
	      if(type==1) 
		{
		  have_eggpoint=1;
  // go back through qlistb to see if any of them were Type 1 and now redundant
		  for(i=0; i<nquarticsb; i++)
		    {
		      if(!qlistbflag[i]) continue;  // i'th is already redundant
		      qlistbflag[i]=(qlistb[i].gettype()!=1);
		      if(!qlistbflag[i]) 
			if(verbose)
			  cout<<"Quartic B #"<<i<<" is now redundant\n";
		    }
		}
	      else 
		{
		  int oldflag=aux_flags[ipivot], newflag=16;
		  if(oldflag==1) newflag=15-pivflag;
		  aux_flags[ipivot]=newflag;
 // go back through qlistb to see if any of them would now be sieved out, 
 // as then they are now redundant
		  long auxpiv = auxs[ipivot];
		  for(i=0; i<nquarticsb; i++)
		    {
		      if(!qlistbflag[i]) continue;  // i'th is already redundant
		      // compute (a,h) of i'th quartic:
		      long a = posmod(qlist[i].geta(),auxpiv);
		      long H = posmod(qlist[i].getH(),auxpiv);
		      
		      // Check if it would now be sieved out:
		      qlistbflag[i] = flags[ipivot][a][H] && newflag;
		      if(!qlistbflag[i]) 
			if(verbose)
			  cout<<"Quartic B #"<<i<<" is now redundant\n";
		    }
		}
	      rank++; selmer_rank++;
	      if(verbose)
		{
		  cout<<"Rank of B=im(eps) increases to "<<rank;
		  if(type==1) cout<<" (The previous point is on the egg)\n";
		  else cout<<" (pivotal prime =" << auxs[ipivot] << ")\n";
		}
 // in this case we do not need to keep the quartic,  so do not increment nquarticsb
	    }
	}  // end of gls case
      else // no rational point was found
	if(els) // have a possible "Selmer point"
	  {
	    if(atype) 
	      {
		n2++;
	      }
	    else
	      {
		qlistbflag[nquarticsb]=1;
		nquarticsb++;
		if(verbose) cout<<" --new (B) #"<<nquarticsb<<endl;
	      }
	  }
      return;
    }
  else
    {
      if (verbose) cout << "--equivalent to ("<<ab<<") #"<<oldnumber<<endl;
    }
} // end of addquartic()


void rank1::getquartics()
{
  nquarticsa = 0; nfirstlota = 0; 
  nquarticsb = 0; nfirstlotb = 0; 
  have_eggpoint = 0;
  rank = selmer_rank = 0;
  ah_count = ah_sieve_0 = ah_sieve_1 = ah_sieve_2 = 0;
  ah_rfail = ah_dfail   = ah_efail   = ah_extra2fail = ah_pass = 0;
  ii=c4; jj=2*c6;   disc = 4*d1728;

  if(div(16,ii)&&div(64,jj)) 
    {
      ii/=16; jj/=64; disc/=4096;
      tr_u/=2; tr_r/=4; tr_s/=2; tr_t/=8;
    }
  if(verbose)
    {
      cout<<"Basic pair: I="<<ii<<", J="<<jj<<endl;
      cout<<"disc="<<disc<<endl;
    }

  xii=I2bigfloat(ii); xjj=I2bigfloat(jj);

  if (div(4,ii)&&div(8,jj)&&div(16,2*ii+jj))
    {
      npairs=1;
      if (verbose) cout<<"One (I,J) pair"<<endl;
    }
  else
    {
      npairs=2;
      if (verbose) cout<<"Two (I,J) pairs"<<endl;
    }

  bigintArray plist0 = getbad_primes(*the_curve);
  // now make sure 2 and 3 are in the list of primes
  long ip=0, jp, np=plist0.length;
  plist = bigintArray(2+np);
  if(plist0.locate(2)<0) plist[ip++]=2;
  if(plist0.locate(3)<0) plist[ip++]=3;
  for(jp=0; jp<np; jp++) plist[ip++]=plist0[jp];
  plist.truncate(ip);

#ifndef NEW_EQUIV
  dlist = sqdivs(disc,plist);
#endif
  threediv = div(3,ii);

  aux_init();
  extra2=0;
  bigcomplex c1(0), c2(-3*xii), c3(xjj);
  cphi = solvecubic( c1, c2, c3);
  Imod2=odd(ii);
  Jmod2=odd(jj);
  if(verbose>1)
    cout<<"Before sorting, phi = "<<cphi[0]<<","<<cphi[1]<<","<<cphi[2]<<endl;
  flag_init();
  getquartics1();
  if (npairs==2) 
    {
      nfirstlota = nquarticsa;
      nfirstlotb = nquarticsb;
      extra2=1;  ii*= 16;  jj*= 64;  disc*= 4096; xii*=16; xjj*=64;
      tr_u*=2; tr_r*=4; tr_s*=2; tr_t*=8;
      Imod2=Jmod2=0;
#ifndef NEW_EQUIV
      dlist = sqdivs(disc,plist);
#endif
      cphi[0]*=4; cphi[1]*=4; cphi[2]*=4;
      getquartics1();
    }
  delete[] cphi;
  if(verbose>1) 
    {
      cout << ah_count      << "\t (a,b,c) triples in search region\n";
      cout << ah_sieve_1    << "\t failed c-divisiblity,\n";
      cout << ah_sieve_2    << "\t failed syzygy sieve,\n";
      cout << ah_sieve_0    << "\t passed sieve.\n";
      cout << ah_rfail      << "\t failed syzygy after sieving,\n";
      cout << ah_dfail      << "\t failed d-integrality,\n";
      cout << ah_efail      << "\t failed e-integrality,\n";
      cout << ah_extra2fail << "\t failed extra-2 divisibility conditions,\n";
      cout << ah_pass       << "\t passed all and produced quartics.\n";
    }
  clear_sieve(); // clears memory allocated by aux_init()
  return;
} // End of getquartics()

void rank1::getquartics1()
{
  if (verbose) 
    cout<<"Looking for quartics with I = "<< ii << ", J = " << jj << endl;

  IJ_curve = Curvedata(0,0,0,-27*ii,-27*jj,0);  // don't minimise

  if (posdisc) 
    {
      gettype(2);  // get type 2s first as they are a subgroup of index 1 or 2
      if(!have_eggpoint) 
	gettype(1);
    }  
  else 
    {
      gettype(3);
    }
}  //  of getquartics1()


void rank1::gettype(int t) // new hybrid version 13/2/96
{
  type=t;
  long a,absa,absa2,b,bb,xc,c,cstep,cmod3;  
  int a_is_odd, b_is_odd, a_div_by_4;
  int a_positive;
  long firsta, lasta, astep, bstep;
  bigint a12, biga8, aa8, bigbb3, I48=48*ii, J64=64*jj;
  bigint cub, rsq, r, rem, h, d, e, ee, bigb, bigbb, bigc, cc, ii_cc;
  bigint iiabcde, jjabcde, qabcde, habcde;
  bigfloat r1, r2, r3, xr, xrsq;
  bigcomplex c1;
  bigfloat phi,phi1,phi2,phi3; 
  bigfloat amax0, amax2, amin2, root27=sqrt(27);;
  bigfloat hmin, hmax, hmin0, hmax0, hmin2, hmax2, htemp;
  bigfloat const1,const7a,const7,const6,const4,const5,const3,const2;

  int extraextra2 = div(64,ii)&&div(128,jj);  // Pascale's extra condition
  cfactor = 1;
  cmod3 = mod(jj,3);

  if (verbose) cout << "Looking for Type " << t << " quartics:\n";

// Set phi to be the real root in type 3, 
// else set the phi_i to be the three real roots 
//   in descending order phi1 > phi2 > phi3:

  switch(type) {
  case 1:
    phi1 = real(cphi[0]);
    phi2 = real(cphi[1]);
    phi3 = real(cphi[2]);
    orderreal(phi1,phi2,phi3);  // decreasing order
    hmax0 = 4*(phi3*phi3-xii)/3;
    break;
  case 2:  // N.B. repetition here to allow types 1 and 2 in either order
    phi1 = real(cphi[0]);
    phi2 = real(cphi[1]);
    phi3 = real(cphi[2]);
    orderreal(phi1,phi2,phi3);  // decreasing order
    hmin0 = 4*(phi2*phi2-xii)/3;
    break;
  case 3:
    // find the real phi and make it phi[2]
    if (is_real(cphi[1])) 
      {
	phi=real(cphi[1]);
	cphi[1]=cphi[2];
	cphi[2]=phi;
      }
    else 
      {
	if (is_real(cphi[2])) 
	  {
	    phi=real(cphi[2]);
	  }
	else 
	  {
	    phi=real(cphi[0]);
	    cphi[0]=cphi[2];
	    cphi[2]=phi;
	  }
      }
    const2 = phi*phi-4*xii;
    const5 = phi*phi-xii;
    const6 = 0;
    if(const2>0) const6 = sqrt(const2);  // const2 must be >0 but avoid rounding problems
    hmin0 = 4*const5/3;
    amax0 = (abs(cphi[0]-cphi[1]) + 2*abs(cphi[0]-cphi[2]))/18;
    if(verbose>1)
      {
	cout<<"After  sorting, phi = "<<cphi[0]<<","
	    << cphi[1]<<","<<cphi[2]
	    << "\nBasic a bound = " << amax0 <<"\n";
      }
  } // end of switch(type)


// Set bounds on a loop:

  switch(type) {
  case 1:
    amin = 1;
    amax =  ROUNDDOWN((phi1-phi3) / 9);
    break;
  case 2:
    amin = ROUNDUP((phi3-phi2)/9);    // negative
    amax = ROUNDDOWN((phi1-phi2)/9);  // positive
    break;
  case 3:
    if(phi>0)
      {
	amax2 = (phi+const6)/6;
	if(amax2>amax0) 
	  {
	    amax2=amax0;
//	    cout<<"New a upper bound worse, not using\n";
	  }
	amax = ROUNDDOWN(amax2);

	amin2= -const6/root27;
	if(amin2<-amax0)
	  {
	    amin2=-amax0;
	    cout<<"New a lower bound worse, not using -- should NOT happen!\n";
	  }
	amin = ROUNDUP(amin2);
      }
    else // phi<0
      {
	amax2= const6/root27;
	if(amax2>amax0)
	  {
	    amax2=amax0;
	    cout<<"New a upper bound worse, not using -- should NOT happen!\n";
	  }
	amax = ROUNDDOWN(amax2);

	amin2 = (phi-const6)/6;
	if(amin2<-amax0)
	  {
	    amin2=-amax0;
//	    cout<<"New a lower bound worse, not using\n";
	  }
	amin = ROUNDUP(amin2);
      }
    if(verbose>1) cout<<"Search range for a: ("<<amin<<","<<amax<<")\n";
  } // end of switch(type)

  for(a_positive=1; a_positive>=0; a_positive--)
    {
      if (a_positive) 
	{
	  firsta=amin; if(firsta<1) firsta=1;
	  lasta =amax;
	  if(firsta>lasta) continue;
	  a=firsta-1;
	  astep=1;
	  if(verbose) 
	    cout << "Trying positive a from " << firsta << " up to " << lasta << endl;
	}
      else  // negative range will be traversed downwards!
	{
	  firsta=amax; if(firsta>-1) firsta=-1;
	  lasta =amin;
	  if(firsta<lasta) continue;
	  a=firsta+1;
	  astep=-1;
	  if(verbose) 
	    cout << "Trying negative a from " << firsta << " down to " << lasta << endl;
	}

  long iaux; 
  long *amodi, *hmodi, *auxi, *hstepmodi, *hscalemodi, *astepmodi;

  int ***flagsi; int **flagai;
  iaux=num_aux; amodi=amod; auxi=auxs;  astepmodi=astepmod;
  while(iaux--) 
    {
      *amodi++     = posmod(a,    *auxi); 
      *astepmodi++ = posmod(astep,*auxi);
      auxi++;
    }

  while(a!=lasta)
    {
      a+=astep;
      for(iaux=0, amodi=amod, auxi=auxs, flagai=flaga, flagsi=flags, 
	  astepmodi=astepmod;
	  iaux<num_aux; 
	  iaux++, amodi++, auxi++, flagai++, flagsi++, astepmodi++) 
	{
	  (*amodi)+=(*astepmodi); 
	  if((*amodi)>=(*auxi)) (*amodi)-=(*auxi);
	  *flagai = (*flagsi)[*amodi];
	}
//
// Tests: not(4|a) if extra2, not(2|a) if extraextra2:
//
      a_is_odd = (a&1);
      a_div_by_4 = !(a&3);
      if ((! (extra2 && (a_div_by_4))) && (! (extraextra2 && !(a_is_odd)))   )
	{
#ifdef SHOW_ABC_RANGES
	  if(verbose)cout<<"a = "<<a<<endl;
#endif
	  efactor = a_is_odd ? 16 : 8;  // only relevant in extra2 case
	  long cfac2 = (extra2?(a_is_odd ? 2 : 4):1);
	  long cfac3 = 3;
	  cstep   = cfac3*cfac2;
	  int b_must_be_odd =  (!extra2) && ((Jmod2 && !a_is_odd) || a_div_by_4);
// latter because else (a,b,c,d,e)==(a/4,b/2,c,2d,4e)
	  bstep = 1;
	  if(extra2) bstep=4; else
	    if(b_must_be_odd) bstep=2;

	  absa=abs(a); absa2=absa<<1;
	  a2=a<<1; a4=a2<<1; a8=a4<<1; a3=a+a2; a6=a3<<1;
	  a12=a8+a4;
	  aa6=a*a6; aa8=a*a8; aa12=2*aa6; aa9=a3*a3;
	  bigfloat xa12=I2bigfloat(a12), xaa8=I2bigfloat(aa8);
	  biga8 = a8; // (as a bigint)
	  bigfloat foura = bigfloat(a4);
	  bigfloat oneover4a = 1/foura;
#ifdef USE_BIGINTS
	  bigint aa=a;  bigint asq=aa*aa; bigint acubed=asq*aa;  
	  bigint asqI48=asq*I48, acubedJ64=acubed*J64;
#else
	  bigfloat asq=a*a; bigfloat acubed=a*asq;
	  bigfloat asqI48=48*asq*xii, acubedJ64=64*acubed*xjj;
#endif
	  
// hstep does not depend on b so can be set here:
	  long hstep = a8*cstep;
	  iaux=num_aux; hstepmodi=hstepmod; auxi=auxs; hscalemodi=hscalemod;
	  if(extra2)
	    while(iaux--) 
	      {
		*hstepmodi++ = posmod(hstep*(*hscalemodi),*auxi); 
		auxi++; hscalemodi++;
	      }
	  else
	    while(iaux--) 
	      {
		*hstepmodi++ = posmod(hstep,*auxi); 
		auxi++;
	      }

// Set up bounds on H

	  switch(type) {
	  case 1:
	    hmin = foura*phi2;
	    hmax = foura*phi3 + hmax0;
	    hmax2 = foura*phi1;
	    if(hmax>hmax2) hmax = hmax2;
	    break;
	  case 2:  //  NB these are really the min/max of sgn(a)*H
	    if(a_positive)
	      {
		hmin = foura*phi2 + hmin0;
		hmax = foura*phi3;
	      }
	    else
	      {
		hmin = foura*phi1;
		hmax = foura*phi2 + hmin0;
	      }
	    break;
	  case 3:
	    {
	      hmax = foura*phi;
	      const3 = 2*sqrt(const2*(4*const5-27*a*a))/3;
	      hmax2 = -hmax/2 + const3;

	      hmin = hmax - hmin0;
	      hmin2 = -hmax/2 - const3;

	      if(hmin2>hmin) hmin=hmin2;
	      if(hmax2<hmax) hmax=hmax2;  // don't move this line up!

	      if(!a_positive) // swap the bounds over:
	      {
		htemp=hmin; hmin=hmax; hmax=htemp;
	      }
	    }
	  }

#ifdef SHOW_ABC_RANGES
	  if(verbose>1)
	    cout<<"(a="<<a<<")\tSearch range for H: ("<<hmin<<","<<hmax<<")\n";
#endif
	  if(a_positive)
	    {
	      if(hmin>hmax) 
		{
		  cout<<"Empty H-range!  hmin = "<<hmin<<", hmax = "<<hmax<<"\n";
		  continue;
		} // skip to next a
	    }
	  else
	    {
	      if(hmax>hmin) 
		{
		  cout<<"Empty H-range!  hmin = "<<hmax<<", hmax = "<<hmin<<"\n";
		  continue;
		} // skip to next a
	    }

// Set up b loop:  first value used is 0 or 1 after incrementing

	  b=b_must_be_odd-bstep;

	  while(b<=(absa2-bstep))
	    {
	      b+=bstep;
#ifdef SHOW_ABC_RANGES
	      if(verbose)cout<<"\tb = "<<b;
#endif
	      b_is_odd = (b&1);
// bstep handles these conditions
/*	      if (
		  ((!Jmod2) || a_is_odd || b_is_odd)
		  && ((!extra2) || !(b&3))
		  && (b_is_odd||(!a_div_by_4))
		  )
*/
		{
		  int doboth = ((0<b) && (b<absa2));
		  b3=3*b; bb=b*b; bb3= 3*bb; bb9= 9*bb; bigbb3=(bigint)bb3;

// set up bounds on c loop: (avoiding rounding error)

		  cmin = ROUNDUP(((hmin + bb3)/a8));
		  cmax = ROUNDDOWN(((hmax + bb3)/a8));

		  if(cmin>cmax+1)
		    cout<<"Empty c-range!  cmin = "<<cmin<<", cmax = "<<cmax<<"\n";

		  while(mod(cmin,3)!=cmod3) 
		    cmin++; // Skip to correct residue mod 3
		  while(cmin%cfac2)       
		    cmin+=3; // Skip to next multiple of cfac2

		  if(cmin>cmax) continue;   // Skip to next b
#ifdef SHOW_ABC_RANGES
		  if(verbose)cout<<":\tcmin = "<<cmin<<", cmax = "<<cmax<<" (cstep = " << cstep << ")\n";
#endif

		  c=cmin-cstep; // So first value used is cmin

		  iaux=num_aux; hmodi=hmod; auxi=auxs; hscalemodi=hscalemod;
		  while(iaux--) 
		    {
		      long aux=(*auxi);
		      long  cmod  = c%aux, 
		           a8mod  = a8%aux, 
		           bb3mod = bb3%aux;
		      long h0 = (((a8mod*cmod)%aux)-bb3mod)%aux; 
		      if(extra2) h0=posmod((*hscalemodi)*h0,aux);
		      else       h0=posmod(h0,aux);
		      *hmodi++ = h0;
		      hscalemodi++;
		      auxi++;
		    }
		  
		  while(c<=cmax-cstep)
		    {
		      ah_count++;
		      c+=cstep;
		      for(iaux=0, hmodi=hmod, hstepmodi=hstepmod, auxi=auxs; 
			  iaux<num_aux; 
			  iaux++, hmodi++, hstepmodi++, auxi++) 
			{
			  (*hmodi) += (*hstepmodi);
			  if((*hmodi)>=(*auxi)) (*hmodi)-=(*auxi);
			}

		      int flagok = (b_is_odd || even(c-Imod2));
		      if(!flagok) {ah_sieve_1++; continue;}
		      
		      ipivot=-1;

		      for(iaux=0, hmodi=hmod, flagai=flaga; 
			  flagok&&(iaux<num_aux); 
			  iaux++, hmodi++, flagai++)
			{
			  int thisflag = (*flagai)[*hmodi];
			  flagok = aux_flags[iaux] & thisflag;
#ifndef NO_PADIC_FILTERING
			  if(flagok&&(iaux>0)&&(thisflag<31)&&(ipivot==-1))
			    {
			      ipivot=iaux; pivflag=thisflag;
			    }
#endif
			}

		      if(!flagok) {ah_sieve_2++; continue;}

// We have an (a,b,c)-triple which passes the sieve test

		      ah_sieve_0++;
		      bigfloat xh=bigfloat(a8)*c-bb3;
		      int ok=1;

// Check that rounding has not put us outside the range:
		      if((a_positive&&((xh>hmax)||(xh<hmin)))
			 ||
			 ((!a_positive)&&((xh>hmin)||(xh<hmax))))
			{
			  if(verbose>1)
			    cout<<"(a,b,c)=("<<a<<","<<b<<","<<c<<"): "
			      <<"H = "<<xh<<" is outside range "
				<<hmin<<"..."<<hmax<<"\n";
			  ok=0;
			}
		      if(!ok) continue;
		      h = biga8*c-bigbb3;

#ifdef USE_BIGINTS
// use bigints from now on
		      cub = h*(h*h-asqI48)+acubedJ64;
		      ok = divides(cub,-27,rsq,rem);
		      if(!ok) 
			{
			  cout<<"cub not divisible by 27 for (a,b,c,h)=("
			    <<a<<","<<b<<","<<c<<","<<h<<")\n";
			  cout<<"cub       = "<<cub<<endl;
			  ah_rfail++;
			  continue;
			}
		      if(!isqrt(rsq,r)) {ah_rfail++; continue;}
		      xr=I2bigfloat(r);
#else
		      xrsq=(xh*(xh*xh-asqI48)+acubedJ64)/(-27);
		      if(!xsqrt(xrsq,xr)) {ah_rfail++; continue;}
#endif
#ifdef DEBUG_AH
		      cout<<"r = "<<r<<" ";
#endif	      
		       bigb = b; bigbb = bigb*bigb;
		       bigc = c; cc = bigc*bigc; ii_cc = ii-cc;  
		       bigfloat xc=c; bigfloat xcc=xc*xc;
#ifdef USE_BIGINTS
		       bigint temp = bigb*(bigbb-a4*bigc);
		       // must compute as bigints
#else
		       bigfloat temp = b*(bb-a4*xc);  
		       // must compute as bigfloats
#endif
// Loop on  sign of b:
		      long ib, sb=b, sb3=b3;
		      for(ib=0; ib<1+doboth; ib++)
			{if(ib) {sb=-sb; sb3=-sb3; temp=-temp;}
#ifdef DEBUG_AH
			 cout<<"\na,b,b3,c,cc,temp = "<<a<<","<<sb<<","
			   <<sb3<<","<<c<<","<<cc<<","<<temp;
#endif
#ifdef USE_BIGINTS
			 ok = divides(r-temp,aa8,d,rem);
#ifdef DEBUG_AH
			 cout<<"; aa8,d,rem = "<<aa8<<","<<d<<","<<rem;
#endif
			 if(!ok) {ah_dfail++; continue;}
#else
			 bigfloat xd = (xr-temp)/xaa8;
			 bigfloat xxd = abs(xd-floor(xd+0.5));
			 if (xxd>abceps) {ah_dfail++; continue;}
#endif
#ifdef USE_BIGINTS
			 ee=ii_cc+sb3*d;
#ifdef DEBUG_AH
			 cout<<"\n ii,b3,d,cc,ee = "<<ii<<","<<sb3<<","
			   <<d<<","<<cc<<","<<ee;
#endif
			 ok = divides(ee,a12,e,rem);
#ifdef DEBUG_AH
			 cout<<"\n ee,a12,e,rem = "<<ee<<","<<a12<<","<<
			   e<<","<<rem;
#endif
			 if(!ok) {ah_efail++; continue;}
#else
			 bigfloat xe = (xii-xcc+sb3*xd)/xa12;
			 bigfloat xxe = abs(xe-floor(xe+0.5));
			 if (xxe>abceps) {ah_efail++; continue;}
			 d = Iround(xd);
			 e = Iround(xe);
			 
#endif
#ifdef DEBUG_AH
			 cout << ":\n [" << a<<","<<sb << "," << c 
			   << "," << d << "," << e << "]\n"; 
#endif
// 
// Now test divisibility conditions in extra2 case:
// (already know 4|b   since 16|h, and not(4|a))
//
			 int skip=0;
			 if(extra2)
			   {
			     skip = !(ndiv(efactor,e)&&ndiv(efactor,a+sb+c+d+e));
			   }
			 if(skip) {ah_extra2fail++; continue;}
			 ah_pass++;
// Now we have a quartic
// Check the invariants are right (for debugging only):
			 iiabcde = II(a,sb,c,d,e);

			 if ( ii != iiabcde )
			   {
			     cout<<"Error: constructed quartic ";
			     
			     cout << "[" << a<<","<<sb << "," << c << "," << d << "," << e << "]"; 
			     cout << " has wrong I-invariant "<<iiabcde<<", not "<<ii<<endl;
			     continue;
			   }
			 jjabcde = JJ(a,sb,c,d,e);
			 if (jj != jjabcde)
			   {
			     cout<<"Error: constructed quartic ";
			     cout << "[" << a<<","<<sb << "," << c << "," << d << "," << e << "]"; 
			     cout << " has wrong J-invariant "<<jjabcde<<", not "<<jj<<endl;
			     //cout << "72ace = " << 72*a*c*e << endl;
			     //cout << "9bcd  = " << sb*(9*c*d) << endl;
			     //cout << "- 27*(a*(d*d) +(b*b*e)) = " << - 27*(a*(d*d) +(bb*e))  << endl;
			     //cout << "-2*c^3 = " << - cc*(2*bigc) << endl;
			     continue;
			   }
// And finally when disc>0 check that the type is correct
			 if(posdisc)
			   {
			     habcde = 8*a*c-3*b*b;
			     qabcde = habcde*habcde-16*a*a*ii;   // =3*Q
			     if(type==1)
			       {
			       if((habcde<0)&&(qabcde>0))
				 {
				   cout<<"Error: constructed quartic ";
				   cout<<"[" << a<<","<<sb << "," << c << "," << d << "," << e << "]"; 
				   cout<<" has type 2, not 1!\n";
				   cout<<"Please report"<<endl;
				   continue;				 
				 }
			     }
			     else // type==2
			       if((habcde>=0)||(qabcde<=0))
				 {
				   cout<<"Error: constructed quartic ";
				   cout<<"[" << a<<","<<sb << "," << c << "," << d << "," << e << "]"; 
				   cout<<" has type 1, not 2!\n";
				   cout<<"Please report"<<endl;
				   continue;				 
				 }
			   }
// Now construct the roots of the quartic and add it to the list.

			 switch(type) {
			 case 1: // no real roots
			   r1 = sqrt((foura*phi1-xh)/3);
			   r2 = sqrt(-(foura*phi2-xh)/3);
			   r3 = sqrt(-(foura*phi3-xh)/3);
			   croots[0]=bigcomplex( r1-sb,  r2-r3) * oneover4a;
			   croots[1]=bigcomplex( r1-sb, -r2+r3) * oneover4a;
			   croots[2]=bigcomplex(-r1-sb,  r2+r3) * oneover4a;
			   croots[3]=bigcomplex(-r1-sb, -r2-r3) * oneover4a;
			   break;
			 case 2: // all real roots
			   r1 = sqrt((foura*phi1-xh)/3);
			   r2 = sqrt((foura*phi2-xh)/3);
			   r3 = sqrt((foura*phi3-xh)/3);
			   croots[0]=bigcomplex( r1+r2-r3-sb) * oneover4a;
			   croots[1]=bigcomplex( r1-r2+r3-sb) * oneover4a;
			   croots[2]=bigcomplex(-r1+r2+r3-sb) * oneover4a;
			   croots[3]=bigcomplex(-r1-r2-r3-sb) * oneover4a;
			   break;
			 case 3: // roots 2,3 are real
			   c1 = sqrt((foura*cphi[0]-xh)/3);
			   r3 = sqrt((foura*phi-xh)/3);
			   if(xr<0) r3=-r3;
			   croots[0]=bigcomplex( r3-sb,  2*imag(c1)) * oneover4a;
			   croots[1]=bigcomplex( r3-sb, -2*imag(c1)) * oneover4a;
			   croots[2]=bigcomplex(-r3-sb  +2*real(c1)) * oneover4a;
			   croots[3]=bigcomplex(-r3-sb  -2*real(c1)) * oneover4a;
			   break;
			 }
			 if((ipivot<0)&&(type>1))
			   qlista[nquarticsa].assign(a,sb,c,d,e,croots,type,ii,jj,disc);
			 else
			   qlistb[nquarticsb].assign(a,sb,c,d,e,croots,type,ii,jj,disc);
			 addquartic();
			 if((type==1)&&(have_eggpoint))
			   {
			     if (verbose) 
			       {
				 cout << "Exiting search for Type 1 quartics after ";
				 cout << "finding one which is globally soluble.\n";
			       }
			     return;
			   }
		       }  // end of b-sign-loop
		    } // end of c loop
		} // end of b conditions
	    } // end of b loop
	} // end of a conditions
    } // end of main a loop
    } // end of loop on sign of a
   if (verbose) cout << "Finished looking for Type " << t << " quartics.\n";
}  // end of gettype()


rank1::rank1(Curvedata* ec, int verb, int treq, long firstlim, long secondlim, long n_aux)
: the_curve(ec), verbose(verb), traceequiv(treq),
  lim1(firstlim), lim2(secondlim),  num_aux(n_aux),
  success(1), rank(0) // default value if failure occurs
{
  if(num_aux==-1) num_aux=DEFAULT_NAUX;
  if(verbose>1) 
    {
      cout << "Using (a,b,c) search with (a,h) sieve and algebraic method\n";
#ifdef USE_BIGINTS
      cout << "(with bigints to solve the syzygy)\n";
#else
      cout << "(with bigfloats to solve the syzygy)\n";
#endif
    }
  
  qlista = new quartic[maxnquartics];
  qlistb = new quartic[maxnquartics];
  qlistbflag = new int[maxnquartics];
  croots = new bigcomplex[4];

  the_curve->getci(c4,c6);
  d1728 = c4*c4*c4-c6*c6;
  if (is_zero(d1728)) {cout<<"Curve is singular\n"; success=0; return;}

  // Set up the transformation [u,r,s,t] from the minimal model to the model
  // [0,0,0,-27*c4,-54*c6]; from these we will later obtain (by simple scaling) 
  // the transformations to the IJ-curve for various I,J
  bigint a1,a2,a3,a4,a6,b2=getb2(*the_curve); 
  the_curve->getai(a1,a2,a3,a4,a6);
  tr_u=6; tr_r=3*b2; tr_s=3*a1; tr_t=108*a3;

  bigintArray ir = Introotscubic( 0,-27*c4,-54*c6);  
  n0=ir.length+1;

  long e0,e1,e2;
  if(!intlog2(n0,e0,0))
    {
      success=0; 
      cout<<"!!! Fatal error: n0 (=#E[2]) not a power of 2\n";
      cout<<"Please inform author by email!\n";
      return;
    }
  
  posdisc = is_positive(d1728);
  pointlist1 = PointArray(maxnquartics);
  pointlist2 = PointArray(2*num_aux);
  npoints1=npoints2=0;
  n1=n2=1;

#ifdef TEST_EQCODE
  cout << "Setting eqplist, length = " << NEQPLIST << "\n";
#endif
  eqplist=longlist(NEQPLIST);
  long jp=0;
  for(primevar pp; jp<NEQPLIST; pp++)
    {
      long p = pp;
      if(ndiv(p,d1728)) eqplist[jp++]=p;
    }
#ifdef TEST_EQCODE
  cout << "eqplist = " << eqplist << "\n";
#endif

  getquartics();
  pointlist1.truncate(npoints1);
  pointlist2.truncate(npoints2);
  delete [] qlista;
  delete [] qlistb;
  delete [] croots;

  // Compute rank/selmer rank from B=im(eps) first:
  // Must count the quartics in qlistb which are still needed
  long n3 = 1;
  for(jp=0; jp<nquarticsb; jp++) n3+=qlistbflag[jp];
  
  long keep_n3 = n3, e3;
  if (!intlog2(n3,e3,1))
    {
      cout<<"\n!!! n3 = "<<keep_n3<<" not a power of 2, rounding up to "<<n3<<"\n";
    }
  selmer_rank = rank + e3;

  if(verbose)
    {
      cout << "rank        contribution from B=im(eps) = " << rank << endl;
      if(e3>0) cout << "Selmer rank contribution from B=im(eps) = " << selmer_rank << endl;
    }
  
  long keep_n1=n1, keep_n2=n2;
  if (!intlog2(n1,e1,1))
    {
      cout<<"\n!!! n1 = "<<keep_n1<<" not a power of 2, rounding up to "<<n1<<"\n";
      cout<<"(Probably due to too small a bound on quartic point search\n";
      cout<<" leading to rational points not being found)\n";
      cout<<" The points listed will be incomplete, but may still\n";
      cout<<" generate a subgroup of finite index.\n";
    }
  if (!intlog2(n2,e2,0))
    {
      success=0; 
      cout<<"\n\n!!! Fatal error: n2 = "<<keep_n2<<" not a power of 2\n";
      cout<<"Please inform author by email!\n";
    }
  
  if(verbose)
    {
      cout << "rank        contribution from A=ker(eps) = " << (e1-e0) << endl;
      if(e2>e1)
	cout << "Selmer rank contribution from A=ker(eps) = " << (e2-e0) << endl;
    }

  n1 <<= rank;
  n2 <<= selmer_rank;
  rank += (e1-e0); 
  selmer_rank += (e2-e0);
  sha2 = (n2/n1);
  certain = (n1==n2);
  int strange = odd(e2-e1);

  if(strange) // Need to make a report even if not verbose
    {
      cout<<"\nWarning: Selmer rank = "<<selmer_rank<<" and program finds \n";
      cout<<"lower bound for rank = "<<rank<<" which differs by an odd\n";
      cout<<"integer from the Selmer rank.   Hence the rank must be 1 more\n";
      cout<<"than reported here.  Try rerunning with a higher bound for\n";
      cout<<"quartic point search.\n";
    }
  
  if (verbose||strange||!certain) 
    {
      cout << "\nSummary of results (all should be powers of 2):\n\n";
      cout<<"n0 = #E(Q)[2]    = "<<n0<<"\n";
      cout<<"n1 = #E(Q)/2E(Q) ";
      if(!certain) cout<<">";
      cout<<"= "<<n1<<"\n";
      cout<<"n2 = #S^(2)(E/Q) = "<<n2<<"\n";
      cout<<"#III(E/Q)[2]     ";
      if(!certain) cout << "<"; 
      cout<<"= "<<sha2<<"\n\n";

      if(certain) 
	cout << "rank " << "= " << rank;
      else 
	cout << rank << " <= rank <= selmer-rank = " << selmer_rank;
      cout << endl << endl; ;

    }  

  if(rank<6) sortpoints();

}  // end of rank1 constructor

void rank1::sortpoints()  // reorder points into increasing height order
{
  long i,j;
  for(i=0; i<npoints1; i++)
    for(j=i+1; j<npoints1; j++)
      if(height(pointlist1[j])<height(pointlist1[i]))
	{
	  Point temp = pointlist1[i]; 
	  pointlist1[i]=pointlist1[j]; 
	  pointlist1[j]=temp;
	}
  for(i=0; i<npoints2; i++)
    for(j=i+1; j<npoints2; j++)
      if(height(pointlist2[j])<height(pointlist2[i]))
	{
	  Point temp = pointlist2[i]; 
	  pointlist2[i]=pointlist2[j]; 
	  pointlist2[j]=temp;
	}
}

void showpoint(Point P, Curvedata* CD=0, const bigint& u=0, const bigint& r=0, 
		                      const bigint& s=0, const bigint& t=0)
{
  Point Q = P;
  if(CD) Q = shift(P,CD,u,r,s,t,1);
  bigfloat h = height(Q);
  cout << Q << ", height = " << h;
  if(!Q.isvalid()) cout << " --warning: NOT on curve!";
  cout << "\n";
}

void rank1::listpoints(Curvedata* CD_orig, const bigint& u, const bigint& r,
		       const bigint& s, const bigint& t)
{
  int explanation_needed = (npoints1>0)&&(npoints2>0);
  if(explanation_needed)
    {
      cout<<"p-adic filtration expresses E(Q)/2E(Q) as a direct sum A+B\n";
      cout<<"where A = E(Q)\\cap\\sum 2E(Q_p) for certain primes p.\n";
      cout<<"We list all nonzero points of A, and generators of B\n";
    }

  if(npoints1>0)
    {
      if(explanation_needed)
	{
	  cout << "Points in A:\n";
	}
      else
	{
	  cout << "Points covering E(Q)/2E(Q):\n";
	}
      for (long i=0; i<npoints1; i++)
	{
	  Point p = pointlist1[i];
	  //	  cout<<p<<" on "<<(Curve)((p.getcurve()))<<endl;
	  cout << "Point "; showpoint(p,CD_orig,u,r,s,t);
	}
    }

  if(npoints2>0)
    {
      if(explanation_needed)
	{
	  cout << "Points generating B:\n";
	}
      else
	{
	  cout << "Points generating E(Q)/2E(Q):\n";
	}
      for (long i=0; i<npoints2; i++)
	{
	  Point p = pointlist2[i];
	  //	  cout<<p<<" on "<<(Curve)((p.getcurve()))<<endl;
	  cout << "Point "; showpoint(p,CD_orig,u,r,s,t);
	}
    }
}

PointArray rank1::getpoints() const
// We construct a set of coset reps for 2E(Q) in E(Q) given
// reps for the subgroup A in pointlist1 and 
// gens for the complementary subgroup B in pointlist2
{
  long np = (1+npoints1) << npoints2;
  PointArray ans(np);
  long j, k, ip=1+npoints1;
  ans[0].init(the_curve,0,1,0);
  for(j=0; j<npoints1; j++) ans[j+1]=pointlist1(j);
  for(j=0; j<npoints2; j++, ip*=2) 
    for(k=0; k<ip; k++)
      {
	ans[ip+k]=ans[k]+pointlist2(j);
      }
  return ans;
}

PointArray rank1::getgens() const
// Returns a set of generators for E(Q) mod 2E(Q)
// (but not necessarily independent)
{
  long j, np = npoints1 + npoints2;
  PointArray ans(np);
  for(j=0; j<npoints2; j++) ans[j]=pointlist2(j);
  for(j=0; j<npoints1; j++) ans[npoints2+j]=pointlist1(j);
  return ans;
}

void rank1::aux_init()  // define  auxiliary moduli and squares
{
  long i, j, a, b;

  auxs = new long[num_aux];
  phimod = new long*[num_aux];
  aux_flags = new int[num_aux];
  squares = new int*[num_aux];
  flags = new int**[num_aux];
  flaga = new int*[num_aux];
  amod = new long[num_aux];
  hmod = new long[num_aux];
  hstepmod = new long[num_aux];
  astepmod = new long[num_aux];
  hscalemod = new long[num_aux];

  auxs[0]=9;  // treated specially
  aux_flags[0]=1;
  i=1;

  // the rest of the auxs must be chosen carefully:  if possible they should
  // be good odd primes p, such that the resolvent cubic is not irreducible mod p.
  // If it has a unique root phi mod p, E(Qp)/2E(Qp) has order 2 and the coset in 
  // which the image of a quartic lies depends on whether it has 0 or 2 roots mod p.
  // If it has 3 roots  mod p, E(Qp)/2E(Qp) has order 4 and the coset in 
  // which the image of a quartic lies depends on whether it has 0 or 4 roots mod p; 
  // if 0, then a further condition determines which non-trivial coset it belongs to

  primevar pr; pr++; pr++;  // skip past 2 and 3

  for(;pr.ok()&&i<num_aux; pr++)
    {
      long p = pr;
      if(div(p,disc)) continue;
      phimod[i] = new long[3];
      long nr;
      long minus3imodp = mod(-3*ii,p);
      long jmodp = mod(jj,p);
      nr = nrootscubic(0,minus3imodp,jmodp,p,phimod[i]);
      if(nr>0) 
	{
	  auxs[i]=p;
	  aux_flags[i] = (nr==1? 8: 1);
	  i++;
	}
    }

  // report on which primes will be used:

  if((verbose>1)&&(num_aux>0)) 
    {
      cout<<"(a,h) sieving using " <<num_aux<<" moduli: \n";
      cout<<"p:\t";
      for(j=0; j<num_aux; j++) cout<<auxs[j]<<"\t";
      cout<<"\n";
      cout<<"code:\t\t";
      for(j=1; j<num_aux; j++) cout<<aux_flags[j]<<"\t";
      cout<<"\n";
      cout<<"phi1:\t\t";
      for(j=1; j<num_aux; j++) cout<<phimod[j][0]<<"\t";
      cout<<"\n";
      cout<<"phi2:\t\t";
      for(j=1; j<num_aux; j++) 
	if(aux_flags[j]==8) cout<<"*\t";
	else 
	  cout<<phimod[j][1]<<"\t";
      cout<<"\n";
      cout<<"phi3:\t\t";
      for(j=1; j<num_aux; j++) 
	if(aux_flags[j]==8) cout<<"*\t";
	else 
	  cout<<phimod[j][2]<<"\t";
      cout<<"\n";
    }

  // initialize flag arrays for squares:

  for (i = 0; i < num_aux; i++)
    {
      long aux = auxs[i];
      long half_aux = ((aux + 1) / 2);
      squares[i] = new int[aux];
      for (j = 0; j < aux; j++)      squares[i][j]=0;
      for (j = 0; j < half_aux; j++) squares[i][posmod( j*j, aux )]=1;

      flags[i] = new int*[aux];
      for(a=0; a<aux; a++)
	flags[i][a] = new int[aux];
    } // end of aux loop

  // initialize scaling factors for use with large I,J pair:
  // NB we use the same sieve for both; (a,h) passes for the larger I,J 
  // iff (a,h/4) passes for the standard I,J.

  for(i=0; i<num_aux; i++)
    {
      hscalemod[i] = invmod(4,auxs[i]);
    }

  if((verbose>1)&&(num_aux>0)) cout<<"finished aux_init()"<<endl;
}

//#define COUNT_CODES

void rank1::flag_init() // set up flag array
{
  if((verbose>1)&&(num_aux>0))  cout<<"starting flag_init()"<<endl;

  int thisflag;
  int ***flagsi=flags;
  int **squaresi=squares;
  long * a4phi= new long[3];
  long * eps  = new long[3];

#ifdef COUNT_CODES
  long * code_count = new long[5]; long icc;
#endif

  for(long i=0; i<num_aux; i++, squaresi++, flagsi++) 
    {
#ifdef COUNT_CODES
      for(icc=0; icc<5; icc++) code_count[icc]=0;
#endif
      
      int case1 = (aux_flags[i]==8);  // phi cubic has 1 root  mod p
      int case2 = !case1;             // phi cubic has 3 roots mod p
      long a, h;
      long aux = auxs[i];
      long aux2 = (i==0 ? 27 : aux);
      long I = I2long(ii%aux2), J = I2long(jj%aux2);
      long I16 = (16*I)%aux2, I48 = (3*I16)%aux2, J64 = (64*J)%aux2;

      int **flagsia = *flagsi;

      for(a=0; a<aux; a++, flagsia++)
	{
	  long a2=(a*a)%aux2;  long a2I48 = (a2*I48)%aux2;  long a2I16 = (a2*I16)%aux2;
	  long a3=(a*a2)%aux2; long a3J64 = (a3*J64)%aux2;
	  if(i>0)
	    {
	      a4phi[0] = (4*a*phimod[i][0])%aux2;
	      if(case2)
		{
		  a4phi[1] = (4*a*phimod[i][1])%aux2;
		  a4phi[2] = (4*a*phimod[i][2])%aux2;
		}
	    }

	  int *flagsiah = *flagsia;

	  for(h=0; h<aux; h++, flagsiah++)
	    {
	      long h2 = (h*h)%aux2;
	      long disc = (((h*((h2-a2I48)%aux2))%aux2)+a3J64)%aux2;
	      if(i==0) // special mod-9 condition to force 27-divisibility
		{
		  *flagsiah = (disc==0);
//		  cout<<"a="<<a<<"\th="<<h<<"\tmod 9 flag = "<<(*flagsiah)<<endl;
		}
	      else
		{
		  disc = posmod(-3*disc,aux2);
		  thisflag = (*squaresi)[disc];
		  if(thisflag) 
		    {
		      // look further to see how many roots mod p an (a,h) quartic
		      // could have.  
		      if(case1)
			{
			  //By choice of auxs there must be 0 or 2 roots, and
			  // the flag is set to 31 if there are 2 roots, else 13
			  thisflag=13;
			  if(disc==0) //must look at Q-seminvariant
			    {
			      long q3 = posmod(3*(h2-a2I16),aux2);
			      if((*squaresi)[q3]) thisflag=31;
			    }
			  else
			    {
			      long z3 = posmod(3*(a4phi[0]-h),aux2);
			      if((*squaresi)[z3]) thisflag=31;
			    }
			}
		      else // case 2
			{
			  // By choice of auxs there must be 0 or 4 roots;
			  // the flag is set to 31 if there are 4 roots, else to
			  // one of 13, 11, 7 depending on which element of
			  // E(Qp)/2E(Qp) the quartic maps to:
			  long iz, z;
			  for(iz=0; iz<3; iz++)
			    {
			      z = posmod(3*(a4phi[iz]-h),aux2);
			      eps[iz] = 2*((*squaresi)[z])-(z==0)-1;
			      // = -1, 0, +1
			    }
// At most one eps is 0 (since E has good reduction at p).
// In this case replace 0 by the value which makes the product +1

			  if(eps[0]==0) {eps[0]=eps[1]*eps[2];}
			  else {
			    if(eps[1]==0) {eps[1]=eps[0]*eps[2];}
			    else {
			      if(eps[2]==0) {eps[2]=eps[0]*eps[1];}
			    }
			  }

			  if(eps[0]) thisflag=(eps[1]? 31: 11);
			  else       thisflag=(eps[1]? 13:  7);
			}
		    }
#ifdef COUNT_CODES
		  if(thisflag==0) code_count[0]++;
		  if(thisflag==31) code_count[1]++;
		  if(thisflag==7) code_count[2]++;
		  if(thisflag==11) code_count[3]++;
		  if(thisflag==13) code_count[4]++;
#endif
		  *flagsiah = thisflag;
		}
	    }
	}
#ifdef COUNT_CODES
      if(i>0)
	{
	  cout << "Code count for p = " << aux << ":\n";
	  cout << 0 << "\t"<< 31 << "\t"<< 7 << "\t"<< 11 << "\t"<< 13 << "\n";
	  for(icc=0; icc<5; icc++) cout<<code_count[icc]<<"\t";
	  cout<<endl;
	}
#endif
    }
  delete [] a4phi; delete [] eps;
#ifdef COUNT_CODES
  delete [] code_count;
#endif

  if((verbose>1)&&(num_aux>0)) 
    cout<<"finished flag_init()"<<endl;

}

void rank1::clear_sieve()  // free memory related to sieve;
{
  for(long i=0; i<num_aux; i++) 
    {
      long aux = auxs[i];
      delete[] squares[i];
      for(long a=0; a<aux; a++)
	{
	  delete[] flags[i][a];
	}
      delete[] flags[i];
    }
  delete[] auxs; delete[] phimod; delete[] squares;
  delete[] aux_flags;
  delete[] flags;  delete[] flaga;
  delete[] amod; delete[] hstepmod; delete[] hscalemod;
  delete[] hmod; delete[] astepmod;
}

// find the number of roots of X^3 + bX^2 + cX + d = 0 (mod p)
// roots are put in r which should be allocated of size 3
int nrootscubic(long b, long c, long d, long p, long* roots)
{
  long r, nr=0;
  int found = 0;
  for (r = 0; (r<p)&&!found ; r++)
    {
      found = (((((r+b)*r+c)*r+d)%p)==0 );
    }
  if (!found) return 0;

  r--;  // because it was incremented one extra time in the loop!
  roots[nr++]=r;
  long e = b + r;
  long f = c + e*r;
  long e0 = (-e*((p+1)/2))%p;
  long dd = posmod(e0*e0-f,p);
  if(legendre(dd,p)==1) 
    { // stupid search is good enough since we only use this for very small p!
      for (r = 1; r<p ; r++)  
	{
	  if((r*r-dd)%p==0) break;
	}
      roots[nr++] = (e0+r)%p;
      roots[nr++] = (e0-r)%p;
    }
  return nr;
}
